'''Command-line :mod:`rejester` management tool.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

This provides a centralized user interface to explore and affect
various parts of the :mod:`rejester` system.  Of note, this is the
principal way to start a worker process to run jobs.  Typical use is:

1. Set up a configuration file containing rejester configuration,
   such as:

    .. code-block:: yaml

        rejester:
          app_name: rejester
          namespace: mynamespace
          registry_addresses: [ "redis.example.com:6379" ]

2. Run some process to generate rejester tasks.

3. Copy the configuration files to other systems as required, and
   run ``rejester_worker -c config.yaml`` to start workers.

4. Run (anywhere, but only once) ``rejester -c config.yaml mode run``
   to actually start execution.

5. Wait for the work units to finish.

6. Run (anywhere, but only once) ``rejester -c config.yaml mode terminate``
   to ask the workers to shut down.

In the configuration, ``namespace`` and ``registry_addresses`` are
required unless passed on the command line.  Additionally,
``registry_addresses`` can be detected from environment variables
``REDIS_PORT_6379_TCP_ADDR`` and ``REDIS_PORT_6379_TCP_PORT``,
which will be set by `Docker <http://www.docker.com/>`_ if a
rejester program is run in a container that is ``--link`` connected
to another container with the name ``redis``.

.. program:: rejester

The :program:`rejester` tool supports the standard
:option:`--config <yakonfig --config>`,
:option:`--dump-config <yakonfig --dump-config>`,
:option:`--verbose <dblogger --verbose>`,
:option:`--quiet <dblogger --quiet>`, and
:option:`--debug <dblogger --debug>` options.  This, and any other tool
that integrates with :mod:`rejester`, supports the following additional
options:

.. option:: --app-name <name>

Provide the application name for database access.  This is combined
with the namespace string.

.. option:: --namespace <name>

Provide the namespace name.  This is qualified by the application
name.  All work units and workers are associated with a single
namespace in a single application.

.. option:: --registry-address <host:port>

Provide the location of a Redis server.

If no further options are given, start an interactive shell to monitor
and control :mod:`rejester`.  Alternatively, a single command can be
given on the command line.  The tool provides the following commands:

.. describe:: summary

    Print a tabular listing of all of the running work specs and
    how many work units are in each state.

.. describe:: load --work-spec file.yaml --work-units file2.json

    Loads a set of work units.  The work spec (``-w``) and work units
    (``-u``) must both be provided as external files.  The work spec
    file is the YAML serialization of a work spec definition; see
    :class:`rejester.TaskMaster` for details of what this looks like.
    The work unit file is a series of JSON records, one to a line,
    each of which is a dictionary of a single ``{"key": {"unit":
    "definition", "dictionary": "values"}}``.

.. describe:: delete

    Deletes the entire namespace.  Prompts for confirmation, unless
    ``-y`` or ``--yes`` is given as an argument.

.. describe:: work_specs

    Prints out the names of all of the work specs.

.. describe:: work_spec --work-spec-name name

    Prints out the definition of a work spec, assuming it has already
    been loaded.  The work spec name may be given with a
    ``--work-spec-name`` or ``-W`` option; or, a ``--work-spec`` or
    ``-w`` option may name a work spec file compatible with the
    ``load`` command.

.. describe:: status --work-spec-name name

    Prints out a summary of the jobs in some work spec.  Provide the
    work spec name the same way as for the ``work_spec`` command.

.. describe:: work_units --work-spec-name name [--status status]

    Prints out a listing of the work units that have not yet completed
    for some work spec.  Provide the work spec name the same way as
    for the ``work_spec`` command.  This includes the work units that
    the ``status`` command would report as "available" or "pending",
    but not other statuses.  If ``-s`` or ``--status`` is given with
    one of the status strings "available", "pending", "blocked",
    "failed", or "finished", only print work units with that status.
    If ``--details`` is given as an argument, print the definition of
    the work unit (and the traceback for failed work units) along with
    its name.

.. describe:: failed --work-spec-name name

    Identical to ``work_units --status failed``.  May be removed at
    a future time.

.. describe:: work_unit --work-spec-name name unitname

    Prints out basic details for a work unit, in any state.

.. describe:: retry --work-spec-name name [--all|unitname...]

    Retry failed work units, removing their traceback and moving them
    back to "available" status.  If ``-a`` or ``--all`` is given,
    retry all failed work units; otherwise, only retry the specific
    work units named on the command line.

.. describe:: clear --work-spec-name name [--status status] [unitname...]

    Remove work units from the system.  With no additional arguments,
    remove all work units from the specified work spec.  If ``-s`` or
    ``--status`` is given, remove only work units with this status;
    see the description of the `work_units` subcommand for possible
    values.  If any ``unitname`` values are given, only remove those
    specific work units, provided they in fact have the specified
    status.

.. describe:: mode [idle|run|terminate]

    With no arguments, print out the current rejester mode; otherwise
    set it.  In "run" mode, workers will start new jobs as they become
    available.  In "idle" mode, workers will not start new jobs but
    also will continue to execute; if the mode is switched back to
    "run" they will start running jobs again.  In "terminate" mode,
    workers will stop execution as soon as they finish their running
    jobs.

.. describe:: run_one

    Get a single task, and run it.

.. describe:: global_lock [--purge]

    The system maintains a global lock to make sequences of database
    requests atomic, but if workers fail, the global lock can be leaked.
    With no arguments, prints the worker ID that owns the global lock
    and details that are known about it.  If ``-p`` or ``--purge`` is
    given, clear out the lock if anybody holds it.

.. describe:: workers

    List all of the known workers.  With ``--all`` include workers
    that haven't checked in recently.  With ``--details`` include all
    known details.

'''
from __future__ import absolute_import
import argparse
import gzip
import json
import logging
import os
import sys
import time

import yaml

import dblogger
import rejester
from rejester.exceptions import NoSuchWorkSpecError, NoSuchWorkUnitError
from rejester._task_master import build_task_master
from rejester.workers import SingleWorker
import yakonfig
from yakonfig.cmd import ArgParseCmd

logger = logging.getLogger(__name__)

def existing_path(string):
    '''"Convert" a string to a string that is a path to an existing file.'''
    if not os.path.exists(string):
        msg = 'path {!r} does not exist'.format(string)
        raise argparse.ArgumentTypeError(msg)
    return string

def existing_path_or_minus(string):
    '''"Convert" a string like :func:`existing_path`, but also accept "-".'''
    if string == '-': return string
    return existing_path(string)

def absolute_path(string):
    '''"Convert" a string to a string that is an absolute existing path.'''
    if not os.path.isabs(string):
        msg = '{!r} is not an absolute path'.format(string)
        raise argparse.ArgumentTypeError(msg)
    if not os.path.exists(os.path.dirname(string)):
        msg = 'path {!r} does not exist'.format(string)
        raise argparse.ArgumentTypeError(msg)
    return string

class Manager(ArgParseCmd):
    def __init__(self):
        ArgParseCmd.__init__(self)
        self.prompt = 'rejester> '
        self._config = None
        self._task_master = None
        self.exitcode = 0

    def precmd(self, line):
        self.exitcode = 0
        return ArgParseCmd.precmd(self, line)

    @property
    def config(self):
        if self._config is None:
            self._config = yakonfig.get_global_config('rejester')
        return self._config

    @property
    def task_master(self):
        """A `TaskMaster` object for manipulating work"""
        if self._task_master is None:
            self._task_master = build_task_master(self.config)
        return self._task_master

    def _add_work_spec_args(self, parser):
        '''Add ``--work-spec`` to an :mod:`argparse` `parser`.'''
        parser.add_argument('-w', '--work-spec', dest='work_spec_path',
                            metavar='FILE', type=existing_path,
                            required=True,
                            help='path to a YAML or JSON file')
    def _get_work_spec(self, args):
        '''Get the contents of the work spec from the arguments.'''
        with open(args.work_spec_path) as f:
            return yaml.load(f)

    def _add_work_spec_name_args(self, parser):
        '''Add either ``--work-spec`` or ``--work-spec-name`` to an
        :mod:`argparse` `parser`.'''
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-w', '--work-spec', dest='work_spec_path',
                           metavar='FILE', type=existing_path,
                           help='path to a YAML or JSON file')
        group.add_argument('-W', '--work-spec-name', metavar='NAME',
                           help='name of a work spec for queries')

    def _get_work_spec_name(self, args):
        '''Get the name of the work spec from the arguments.

        This assumes :meth:`_add_work_spec_name_args` has been called,
        but will also work if just :meth:`_add_work_spec_args` was
        called instead.

        '''
        if getattr(args, 'work_spec_name', None):
            return args.work_spec_name
        return self._get_work_spec(args)['name']

    def _read_work_units_file(self, work_units_fh):
        '''work_units_fh is iterable on lines (could actually be ['{}',...])
        '''
        count = 0
        for line in work_units_fh:
            if not line:
                continue
            line = line.strip()
            if not line:
                continue
            if line[0] == '#':
                continue
            try:
                count += 1
                work_unit = json.loads(line)
                #work_units.update(work_unit)
                for k,v in work_unit.iteritems():
                    yield k,v
            except:
                logger.error('failed handling work_unit on line %s: %r', count, line, exc_info=True)
                raise

    def _work_units_fh_from_path(self, work_units_path):
        work_units_fh = None
        if work_units_path == '-':
            work_units_fh = sys.stdin
        elif work_units_path is not None:
            if work_units_path.endswith('.gz'):
                work_units_fh = gzip.open(work_units_path)
            else:
                work_units_fh = open(work_units_path)
        return work_units_fh

    def args_load(self, parser):
        self._add_work_spec_args(parser)
        parser.add_argument('-n', '--nice', default=0, type=int,
                            help='specify a nice level for these jobs')
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-u', '--work-units', metavar='FILE',
                           dest='work_units_path',
                           type=existing_path_or_minus,
                           help='path to file with one JSON record per line')
        group.add_argument('--no-work', default=False, action='store_true',
                           help='set no work units, just the work spec')

    def do_load(self, args):
        '''loads work_units into a namespace for a given work_spec

        --work-units file is JSON, one work unit per line, each line a
        JSON dict with the one key being the work unit key and the
        value being a data dict for the work unit.

        '''
        work_spec = self._get_work_spec(args)
        work_spec['nice'] = args.nice
        self.task_master.set_work_spec(work_spec)

        if args.no_work:
            work_units_fh = []  # it just has to be an iterable with no lines
        else:
            work_units_fh = self._work_units_fh_from_path(args.work_units_path)
            if work_units_fh is None:
                raise RuntimeError('need -u/--work-units or --no-work')
        self.stdout.write('loading work units from {!r}\n'
                          .format(work_units_fh))
        work_units = {k:v for k,v in self._read_work_units_file(work_units_fh)}

        if work_units:
            self.stdout.write('pushing work units\n')
            self.task_master.add_work_units(work_spec['name'], work_units.items())
            self.stdout.write('finished writing {} work units to work_spec={!r}\n'
                              .format(len(work_units), work_spec['name']))
        else:
            self.stdout.write('no work units. done.\n')

    def args_addwork(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('-u', '--work-units', metavar='FILE',
                            dest='work_units_path',
                            type=existing_path_or_minus,
                            help='path to file with one JSON record per line')

    def do_addwork(self, args):
        work_spec_name = self._get_work_spec_name(args)
        work_units_fh = self._work_units_fh_from_path(args.work_units_path)
        work_units = [kv for kv in self._read_work_units_file(work_units_fh)]
        if work_units_fh is None:
            raise RuntimeError('need -u/--work-units')

        self.task_master.add_work_units(work_spec_name, work_units)

    def args_delete(self, parser):
        parser.add_argument('-y', '--yes', default=False, action='store_true',
                            dest='assume_yes',
                            help='assume "yes" and require no input for '
                            'confirmation questions.')
    def do_delete(self, args):
        '''delete the entire contents of the current namespace'''
        namespace = self.config['namespace']
        if not args.assume_yes:
            response = raw_input('Delete everything in {!r}?  Enter namespace: '
                                 .format(namespace))
            if response != namespace:
                self.stdout.write('not deleting anything\n')
                return
        self.stdout.write('deleting namespace {!r}\n'.format(namespace))
        self.task_master.registry.delete_namespace()

    def args_work_specs(self, parser):
        pass
    def do_work_specs(self, args):
        '''print the names of all of the work specs'''
        work_spec_names = [x['name'] for x in self.task_master.iter_work_specs()]
        work_spec_names.sort()
        for name in work_spec_names:
            self.stdout.write('{}\n'.format(name))

    def args_work_spec(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('--json', default=False, action='store_true',
                            help='write output in json')
        parser.add_argument('--yaml', default=False, action='store_true',
                            help='write output in yaml (default)')
    def do_work_spec(self, args):
        '''dump the contents of an existing work spec'''
        work_spec_name = self._get_work_spec_name(args)
        spec = self.task_master.get_work_spec(work_spec_name)
        if args.json:
            self.stdout.write(json.dumps(spec, indent=4, sort_keys=True) + '\n')
        else:
            yaml.safe_dump(spec, self.stdout)

    def args_status(self, parser):
        self._add_work_spec_name_args(parser)
    def do_status(self, args):
        '''print the number of work units in an existing work spec'''
        work_spec_name = self._get_work_spec_name(args)
        status = self.task_master.status(work_spec_name)
        self.stdout.write(json.dumps(status, indent=4, sort_keys=True) + '\n')

    def args_summary(self, parser):
        pass
    def do_summary(self, args):
        '''print a summary of running rejester work'''
        self.stdout.write('Work spec               Avail  Pending  Blocked'
                          '   Failed Finished    Total\n')
        self.stdout.write('==================== ======== ======== ========'
                          ' ======== ======== ========\n')
        for ws in self.task_master.iter_work_specs():
            name = ws['name']
            status = self.task_master.status(name)
            self.stdout.write('{0:20s} {1[num_available]:8d} '
                              '{1[num_pending]:8d} {1[num_blocked]:8d} '
                              '{1[num_failed]:8d} {1[num_finished]:8d} '
                              '{1[num_tasks]:8d}\n'.format(name, status))

    def args_work_units(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('-n', '--limit', type=int, metavar='N',
                            help='only print N work units')
        parser.add_argument('-s', '--status',
                            choices=['available', 'pending', 'blocked',
                                     'finished', 'failed'],
                            help='print work units in STATUS')
        parser.add_argument('--details', action='store_true',
                            help='also print the contents of the work units')
    def do_work_units(self, args):
        '''list work units that have not yet completed'''
        work_spec_name = self._get_work_spec_name(args)
        if args.status:
            status = args.status.upper()
            statusi = getattr(self.task_master, status, None)
            if statusi is None:
                self.stdout.write('unknown status {!r}\n'.format(args.status))
                return
        else:
            statusi = None
        work_units = dict(self.task_master.get_work_units(work_spec_name, state=statusi, limit=args.limit))
        work_unit_names = sorted(work_units.keys())
        if args.limit: work_unit_names = work_unit_names[:args.limit]
        for k in work_unit_names:
            if args.details:
                tback = work_units[k].get('traceback', '')
                if tback:
                    tback += '\n'
                    work_units[k]['traceback'] = 'displayed below'
                self.stdout.write('{!r}: {}\n{}'
                    .format(k, json.dumps(work_units[k], indent=4, sort_keys=True),
                            tback))
            else:
                self.stdout.write('{}\n'.format(k))

    def args_failed(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('-n', '--limit', type=int, metavar='N',
                            help='only print N work units')
        parser.add_argument('--details', action='store_true',
                            help='also print the contents of the work units')
    def do_failed(self, args):
        '''list failed work units (deprecated)'''
        args.status = 'failed'
        return self.do_work_units(args)

    def args_work_unit(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('unit', nargs='*',
                            help='work unit name(s)')
    def do_work_unit(self, args):
        '''print basic details about work units'''
        work_spec_name = self._get_work_spec_name(args)
        for work_unit_name in args.unit:
            status = self.task_master.get_work_unit_status(work_spec_name,
                                                           work_unit_name)
            self.stdout.write('{} ({!r})\n'
                              .format(work_unit_name, status['status']))
            if 'expiration' in status:
                when = time.ctime(status['expiration'])
                if status == 'available':
                    if status['expiration'] == 0:
                        self.stdout.write('  Never scheduled\n')
                    else:
                        self.stdout.write('  Available since: {}\n'
                                          .format(when))
                else:
                    self.stdout.write('  Expires: {}\n'.format(when))
            if 'worker_id' in status:
                try:
                    heartbeat = self.task_master.get_heartbeat(status['worker_id'])
                except:
                    heartbeat = None
                if heartbeat:
                    hostname = (heartbeat.get('fqdn', None) or
                                heartbeat.get('hostname', None) or
                                '')
                    ipaddrs = ', '.join(heartbeat.get('ipaddrs', ()))
                    if hostname and ipaddrs:
                        summary = '{} on {}'.format(hostname, ipaddrs)
                    else:
                        summary = hostname + ipaddrs
                else:
                    summary = 'No information'
                self.stdout.write('  Worker: {} ({})\n'.format(
                    status['worker_id'], summary))
            if 'traceback' in status:
                self.stdout.write('  Traceback:\n{}\n'.format(
                    status['traceback']))
            if 'depends_on' in status:
                self.stdout.write('  Depends on:\n')
                for what in status['depends_on']:
                    self.stdout.write('    {!r}\n'.format(what))

    def args_retry(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('-a', '--all', action='store_true',
                            help='retry all failed jobs')
        parser.add_argument('unit', nargs='*',
                            help='work unit name(s) to retry')
    def do_retry(self, args):
        '''retry a specific failed job'''
        work_spec_name = self._get_work_spec_name(args)
        retried = 0
        complained = False
        try:
            if args.all:
                while True:
                    units = self.task_master.get_work_units(
                        work_spec_name, limit=1000,
                        state=self.task_master.FAILED)
                    units = [u[0] for u in units]  # just need wu key
                    if not units: break
                    try:
                        self.task_master.retry(work_spec_name, *units)
                        retried += len(units)
                    except NoSuchWorkUnitError, e:
                        # Because of this sequence, this probably means
                        # something else retried the work unit.  If we
                        # try again, we shouldn't see it in the failed
                        # list...so whatever
                        pass
            else:
                units = args.unit
                try:
                    self.task_master.retry(work_spec_name, *units)
                    retried += len(units)
                except NoSuchWorkUnitError, e:
                    unit = e.work_unit_name
                    self.stdout.write('No such failed work unit {!r}.\n'
                                      .format(unit))
                    complained = True
                    units.remove(unit)
                    # and try again
        except NoSuchWorkSpecError, e:
            # NB: you are not guaranteed to get this, especially with --all
            self.stdout.write('Invalid work spec {!r}.\n'
                              .format(work_spec_name))
            return
        if retried == 0 and not complained:
            self.stdout.write('Nothing to do.\n')
        elif retried == 1:
            self.stdout.write('Retried {} work unit.\n'.format(retried))
        elif retried > 1:
            self.stdout.write('Retried {} work units.\n'.format(retried))

    def args_clear(self, parser):
        self._add_work_spec_name_args(parser)
        parser.add_argument('-s', '--status',
                            choices=['available', 'pending', 'blocked',
                                     'finished', 'failed'],
                            help='print work units in STATUS')
        parser.add_argument('unit', nargs='*',
                            help='work unit name(s) to remove')
    def do_clear(self, args):
        '''remove work units from a work spec'''
        # Which units?
        work_spec_name = self._get_work_spec_name(args)
        units = args.unit or None
        # What to do?
        count = 0
        if args.status is None:
            all = units is None
            count += self.task_master.del_work_units(work_spec_name, work_unit_keys=units, all=all)
        elif args.status == 'available':
            count += self.task_master.del_work_units(
                work_spec_name, work_unit_keys=units, state=self.task_master.AVAILABLE)
        elif args.status == 'pending':
            count += self.task_master.del_work_units(
                work_spec_name, work_unit_keys=units, state=self.task_master.PENDING)
        elif args.status == 'blocked':
            count += self.task_master.del_work_units(
                work_spec_name, work_unit_keys=units, state=self.task_master.BLOCKED)
        elif args.status == 'finished':
            count += self.task_master.del_work_units(
                work_spec_name, work_unit_keys=units, state=self.task_master.FINISHED)
        elif args.status == 'failed':
            count += self.task_master.del_work_units(
                work_spec_name, work_unit_keys=units, state=self.task_master.FAILED)
        self.stdout.write('Removed {} work units.\n'.format(count))

    def args_mode(self, parser):
        parser.add_argument('mode', choices=['idle', 'run', 'terminate'],
                            nargs='?',
                            help='set rejester worker mode to MODE')
    def do_mode(self, args):
        '''get or set the global rejester worker mode'''
        if args.mode:
            mode = { 'idle': self.task_master.IDLE,
                     'run': self.task_master.RUN,
                     'terminate': self.task_master.TERMINATE }[args.mode]
            self.task_master.set_mode(mode)
            self.stdout.write('set mode to {!r}\n'.format(args.mode))
        else:
            mode = self.task_master.get_mode()
            self.stdout.write('{!s}\n'.format(mode))

    def args_global_lock(self, parser):
        parser.add_argument('--purge', '-p', action='store_true',
                            help='forcibly clear the lock')
    def do_global_lock(self, args):
        '''read (or clear) the global lock'''
        if args.purge:
            self.task_master.registry.force_clear_lock()
        else:
            owner = self.task_master.registry.read_lock()
            if owner:
                heartbeat = self.task_master.get_heartbeat(owner)
                if 'hostname' in heartbeat:
                    self.stdout.write('{} ({})\n'.format(owner,
                                                         heartbeat['hostname']))
                else:
                    self.stdout.write('{}\n'.format(owner))
            else:
                self.stdout.write('(unlocked)\n')

    def args_workers(self, parser):
        parser.add_argument('--all', action='store_true',
                            help='list all workers (even dead ones)')
        parser.add_argument('--details', action='store_true',
                            help='include more details if available')
    def do_workers(self, args):
        '''list all known workers'''
        workers = self.task_master.workers(alive=not args.all)
        for k in sorted(workers.iterkeys()):
            self.stdout.write('{} ({})\n'.format(k, workers[k]))
            if args.details:
                heartbeat = self.task_master.get_heartbeat(k)
                for hk, hv in heartbeat.iteritems():
                    self.stdout.write('  {}: {}\n'.format(hk, hv))

    def args_run_one(self, parser):
        pass
    def do_run_one(self, args):
        '''run a single job'''
        worker = SingleWorker(self.config, task_master=self.task_master)
        worker.register()
        rc = False
        try:
            rc = worker.run()
        finally:
            worker.unregister()
        if not rc:
            self.exitcode = 2

def main():
    parser = argparse.ArgumentParser(
        description='manage the rejester distributed work system')
    mgr = Manager()
    mgr.add_arguments(parser)
    args = yakonfig.parse_args(parser, [yakonfig, dblogger, rejester])
    mgr.main(args)
    sys.exit(mgr.exitcode)

if __name__ == '__main__':
    main()



'''Rejester workers.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

The standard worker infrastructure in the classes below calls
:meth:`rejester.WorkUnit.run` on individual work units as they become
available.  In normal use, a caller will use
:meth:`rejester.TaskMaster.update_bundle` to submit jobs, then expect
an external caller to run ``rejester run_worker``, which will create a
:class:`MultiWorker` object that runs those jobs.

Other implementation strategies are definitely possible.  The
:class:`SingleWorker` class here will run exactly one job when
invoked.  It is also possible for a program that intends to do some
work, possibly even in parallel, but wants to depend on rejester for
queueing, to call :meth:`rejester.TaskMaster.get_work` itself and do
work based on whatever information is in the work spec; that would not
use this worker infrastructure at all.

.. autoclass:: SingleWorker
    :members:
    :show-inheritance:

.. autoclass:: MultiWorker
    :members:
    :show-inheritance:

.. autoclass:: HeadlessWorker
    :members:
    :show-inheritance:

.. autoclass:: ForkWorker
    :members:
    :show-inheritance:

.. autofunction:: run_worker

'''
from __future__ import absolute_import
import errno
import logging
import multiprocessing
from operator import itemgetter
import os
import psutil
import random
import Queue
import signal
import struct
import sys
import threading
import time
import traceback

from setproctitle import setproctitle

import dblogger
import rejester
from rejester._task_master import build_task_master
from rejester import Worker
from rejester.exceptions import LostLease
from rejester._registry import nice_identifier
import yakonfig

logger = logging.getLogger(__name__)


def test_work_program(work_unit):
    # just to show that this works, we get the config from the data
    # and *reconnect* to the registry with a second instances instead
    # of using work_unit.registry
    config = work_unit.data['config']
    sleeptime = float(work_unit.data.get('sleep', 9.0))
    task_master = build_task_master(config)
    logger.info('executing work_unit %r ... %s', work_unit.key, sleeptime)
    time.sleep(sleeptime)  # pretend to work
    logger.info('finished %r' % work_unit)


def run_worker(worker_class, *args, **kwargs):
    '''Bridge function to run a worker under :mod:`multiprocessing`.

    The :mod:`multiprocessing` module cannot
    :meth:`~multiprocessing.Pool.apply_async` to a class constructor,
    even if the ``__init__`` calls ``.run()``, so this simple wrapper
    calls ``worker_class(*args, **kwargs)`` and logs any exceptions
    before re-raising them.

    This is usually only used to create a :class:`HeadlessWorker`, but
    it does run through the complete
    :meth:`~rejester.Worker.register`, :meth:`~rejester.Worker.run`,
    :meth:`~rejester.Worker.unregister` sequence with some logging
    on worker-level failures.

    '''
    try:
        worker = worker_class(*args, **kwargs)
    except Exception:
        logger.critical('failed to create worker {!r}'.format(worker_class),
                        exc_info=True)
        raise
    # A note on style here:
    #
    # If this runs ForkWorker, ForkWorker will os.fork() LoopWorker
    # (or SingleWorker) children, and the child will have this in its
    # call stack.  Eventually the child will sys.exit(), which raises
    # SystemExit, which is an exception that will trickle back through
    # here.  If there is a try:...finally: block, the finally: block
    # will execute on every child exit.  except Exception: won't run
    # on SystemExit.
    try:
        worker.register()
        worker.run()
        worker.unregister()
    except Exception:
        logger.error('worker {!r} died'.format(worker_class), exc_info=True)
        try:
            worker.unregister()
        except:
            pass
        raise


class HeadlessWorker(Worker):
    '''Child worker to do work under :mod:`multiprocessing`.

    The :meth:`run` method expects to run a single
    :class:`~rejester.WorkUnit`, which it will receive from its
    parent :class:`MultiWorker`.  This class expects to be the only
    thing run in a :mod:`multiprocessing` child process.

    '''

    def __init__(self, config, worker_id, work_spec_name, work_unit_key):
        # Do a complete reset of logging right now before we do anything else.
        #
        # multiprocessing.Pool is super super asynchronous: when you
        # apply_async() your job description goes into a queue, which
        # a thread moves to another queue, a second thread tries every
        # 0.1s to make sure the subprocesses exist, and the subprocess
        # actually pulls the job off the queue.  If your main thread is
        # doing something when that 0.1s timer fires, it's possible that
        # the os.fork()ed child is actually forked holding some lock from
        # the parent.
        #
        # See:  http://bugs.python.org/issue6721
        #
        # logging seems to be the most prominent thing that causes
        # trouble here.  There is both a global logging._lock,
        # plus every logging.Handler instance has a lock.  If we just
        # clean up these locks we'll be good.
        logging._lock = threading.RLock()
        for handler in logging._handlers.itervalues():
            handler.createLock()

        # Now go on as normal
        super(HeadlessWorker, self).__init__(config)
        for sig_num in [signal.SIGTERM, signal.SIGHUP, signal.SIGABRT]:
            signal.signal(sig_num, self.terminate)
        self.work_unit = self.task_master.get_assigned_work_unit(
            worker_id, work_spec_name, work_unit_key)
        # carry this to overwrite self.worker_id after .register()
        self._pre_assigned_worker_id = worker_id

    def register(self):
        super(HeadlessWorker, self).register()
        self.worker_id = self._pre_assigned_worker_id

    def run(self):
        self.work_unit.run()

    def terminate(self, sig_num, frame):
        logger.info('received %d, ending work unit', sig_num)
        self.work_unit.terminate()
        sys.exit()


class MultiWorker(Worker):
    '''Parent worker that runs multiple jobs continuously.

    This uses :mod:`multiprocessing` to run one child
    :class:`HeadlessWorker` per core on the system, and averages
    system memory to report `available_gb`.  This class manages the
    :class:`~rejester.TaskMaster` interactions and sends
    :class:`~rejester.WorkUnit` instances to its managed child
    processes.

    This class is normally invoked from the command line by
    running ``rejester run_worker``, which runs this class as a
    daemon process.

    Instances of this class, running across many machines in a
    cluster, are controlled by :meth:`rejester.TaskMaster.get_mode`.
    The :meth:`run` method will exit if the current mode is
    :attr:`~rejester.TaskMaster.TERMINATE`.  If the mode is
    :attr:`~rejester.TaskMaster.IDLE` then the worker will stay
    running but will not start new jobs.  New jobs will be started
    only when the mode becomes :attr:`~rejester.TaskMaster.RUN`.  The
    system defaults to :attr:`~rejester.TaskMaster.IDLE` state, but if
    workers exit immediately, it may be because the mode has been left
    at :attr:`~rejester.TaskMaster.TERMINATE` from a previous
    execution.

    If `tasks_per_cpu` is set in the configuration block for rejester,
    then that many child process will be launched for each CPU on the
    machine.

    '''
    def __init__(self, config):
        super(MultiWorker, self).__init__(config)
        self._event_queue = multiprocessing.Queue()
        self._mode = None
        self.pool = None
        logger.debug('MultiWorker initialized')

    _available_gb = None

    @classmethod
    def available_gb(cls):
        if cls._available_gb is None:
            mem = psutil.virtual_memory()
            cls._available_gb = float(mem.available) / (multiprocessing.cpu_count() * 2**30)
        return cls._available_gb

    def _finish_callback(self, *args):
        # We don't actually get anything useful from the work call, so
        # just post an event that causes us to wake up and poll all
        # the slots.
        self._event_queue.put(True)

    def _poll_async_result(self, async_result, work_unit, do_update=True):
        if async_result is None:
            return
        assert work_unit is not None
        if not async_result.ready():
            if do_update:
                logger.debug('not ready %r, update', work_unit.key)
                work_unit.update()
            return
        try:
            async_result.get(0)
        except multiprocessing.TimeoutError:
            if do_update:
                logger.debug('get timeout update %r', work_unit.key)
                work_unit.update()
            return
        except Exception, exc:
            logger.error('trapped child exception', exc_info=True)
            work_unit.fail(exc)
        else:
            ## if it gets here, slot should always be finished
            assert async_result.ready()
            work_unit.finish()
        ## either failed or finished
        assert work_unit.failed or work_unit.finished

    def _get_and_start_work(self):
        "return (async_result, work_unit) or (None, None)"
        worker_id = nice_identifier()
        work_unit = self.task_master.get_work(worker_id, available_gb=self.available_gb())
        if work_unit is None:
            return None, None
        async_result = self.pool.apply_async(
            run_worker,
            (HeadlessWorker, self.task_master.registry.config,
             worker_id,
             work_unit.work_spec_name,
             work_unit.key),
            callback=self._finish_callback)
        return async_result, work_unit

    def _poll_slots(self, slots, mode=None, do_update=False):
        hasWork = True
        for i in xrange(len(slots)):
            async_result, work_unit = slots[i]
            if (async_result is not None) and (work_unit is not None):
                self._poll_async_result(async_result, work_unit, do_update=do_update)
                if work_unit.failed or work_unit.finished:
                    slots[i] = (None, None)
            if (slots[i][0] is None) and (mode == self.task_master.RUN):
                if hasWork:
                    slots[i] = self._get_and_start_work()
                    if slots[i][0] is None:
                        # If we fail to get work, don't hammer the
                        # taskmaster with requests for work. Wait
                        # until after a sleep and the next _poll_slots
                        # cycle.
                        hasWork = False

    def run(self):
        '''Fetch and dispatch jobs as long as the system is running.

        This periodically checks the :class:`rejester.TaskMaster` mode
        and asks it for more work.  It will normally run forever in a
        loop until the mode becomes
        :attr:`~rejester.TaskMaster.TERMINATE`, at which point it
        waits for all outstanding jobs to finish and exits.

        This will :func:`~rejester.Worker.heartbeat` and check for new
        work whenever a job finishes, or otherwise on a random
        interval between 1 and 5 seconds.

        '''

        tm = self.task_master
        num_workers = multiprocessing.cpu_count()
        if 'tasks_per_cpu' in self.config:
            num_workers *= self.config.get('tasks_per_cpu') or 1
        if self.pool is None:
            self.pool = multiprocessing.Pool(num_workers, maxtasksperchild=1)
        ## slots is a fixed-length list of [AsyncRsults, WorkUnit]
        slots = [[None, None]] * num_workers
        logger.info('MultiWorker starting with %s workers', num_workers)
        min_loop_time = 2.0
        lastFullPoll = time.time()
        while True:
            mode = self.heartbeat()
            if mode != self._mode:
                logger.info('worker {} changed to mode {}'
                            .format(self.worker_id, mode))
                self._mode = mode
            now = time.time()
            should_update = (now - lastFullPoll) > min_loop_time
            self._poll_slots(slots, mode=mode, do_update=should_update)
            if should_update:
                lastFullPoll = now

            if mode == tm.TERMINATE:
                num_waiting = sum(map(int, map(bool, map(itemgetter(0), slots))))
                if num_waiting == 0:
                    logger.info('MultiWorker all children finished')
                    break
                else:
                    logger.info('MultiWorker waiting for %d children to finish', num_waiting)

            sleepsecs = random.uniform(1,5)
            sleepstart = time.time()
            try:
                self._event_queue.get(block=True, timeout=sleepsecs)
                logger.debug('woken by event looptime=%s sleeptime=%s', sleepstart - now, time.time() - sleepstart)
            except Queue.Empty, empty:
                logger.debug('queue timed out. be exhausting, looptime=%s sleeptime=%s', sleepstart - now, time.time() - sleepstart)
                # it's cool, timed out, do the loop of checks and stuff.

        logger.info('MultiWorker exiting')


class SingleWorker(Worker):
    '''Worker that runs exactly one job when called.

    This is used by the :meth:`rejester.run.Manager.do_run_one`
    command to run a single job; that just calls :meth:`run`.  This
    is also invoked as the child process by :class:`ForkWorker`,
    which calls :meth:`as_child`.

    '''
    def run(self, set_title=False):
        '''Do some work.

        The standard implementation here calls :meth:`run_one`.

        :param set_title: if true, set the process's title with the
          work unit name
        :return: :const:`True` if there was a job (even if it failed)

        '''
        return self.run_one(set_title)

    def run_one(self, set_title=False):
        '''Get exactly one job, run it, and return.

        Does nothing (but returns :const:`False`) if there is no work
        to do.  Ignores the global mode; this will do work even
        if :func:`rejester.TaskMaster.get_mode` returns
        :attr:`~rejester.TaskMaster.TERMINATE`.

        :param set_title: if true, set the process's title with the
          work unit name
        :return: :const:`True` if there was a job (even if it failed)

        '''
        available_gb = MultiWorker.available_gb()
        unit = self.task_master.get_work(self.worker_id, available_gb)
        if unit is None:
            logger.info('No work to do; stopping.')
            return False
        try:
            if set_title:
                setproctitle('rejester worker {} {}'
                             .format(unit.work_spec_name, unit.key))
            unit.run()
            unit.finish()
        except LostLease, e:
            # We don't own the unit any more so don't try to report on it
            pass
        except Exception, e:
            unit.fail(e)
        return True

    #: Exit code from :meth:`as_child` if it ran a work unit (maybe
    #: unsuccessfully).
    EXIT_SUCCESS = 0
    #: Exit code from :meth:`as_child` if there was a failure getting
    #: the work unit.
    EXIT_EXCEPTION = 1
    #: Exit code from :meth:`as_child` if there was no work to do.
    EXIT_BORED = 2

    @classmethod
    def as_child(cls, global_config, parent=None):
        '''Run a single job in a child process.

        This method never returns; it always calls :func:`sys.exit`
        with an error code that says what it did.

        '''
        try:
            setproctitle('rejester worker')
            random.seed()  # otherwise everyone inherits the same seed
            yakonfig.set_default_config([yakonfig, dblogger, rejester],
                                        config=global_config)
            worker = cls(yakonfig.get_global_config(rejester.config_name))
            worker.register(parent=parent)
            did_work = worker.run(set_title=True)
            worker.unregister()
            if did_work:
                sys.exit(cls.EXIT_SUCCESS)
            else:
                sys.exit(cls.EXIT_BORED)
        except Exception, e:
            # There's some off chance we have logging.
            # You will be here if redis is down, for instance,
            # and the yakonfig dblogger setup runs but then
            # the get_work call fails with an exception.
            if len(logging.root.handlers) > 0:
                logger.critical('failed to do any work', exc_info=e)
            sys.exit(cls.EXIT_EXCEPTION)


class LoopWorker(SingleWorker):
    '''Worker that runs jobs for a fixed length of time.

    This does jobs as :meth:`SingleWorker.run_one`.  However,
    the worker itself is configured with a maximum lifetime, and
    if a job finishes before the lifetime has passed, this worker
    will try to do another job.

    '''
    def run(self, set_title=False):
        '''Do some work.

        The standard implementation here calls :meth:`run_loop`.

        :param set_title: if true, set the process's title with the
          work unit name
        :return: :const:`True` if there was a job (even if it failed)

        '''
        return self.run_loop(set_title)

    def run_loop(self, set_title=False):
        # Find our deadline from the global configuration.
        # We don't really love this, but it's data we're already
        # passing across process boundaries.
        now = time.time()
        try:
            duration = yakonfig.get_global_config('rejester', 'fork_worker',
                                                  'child_lifetime')
        except KeyError:
            duration = 10
        deadline = now + duration

        while time.time() < deadline:
            ret = self.run_one(set_title)
            if not ret:
                break

        return ret


class ForkWorker(Worker):
    '''Parent worker that runs multiple jobs concurrently.

    This manages a series of child processes, each of which runs
    a :class:`SingleWorker`.  It runs as long as the global rejester
    state is not :class:`rejester.TaskMaster.TERMINATE`.

    This takes some additional optional configuration options.  A
    typical configuration will look like:

    .. code-block:: yaml

        rejester:
          # required rejester configuration
          registry_addresses: [ 'redis.example.com:6379' ]
          app_name: rejester
          namespace: namespace

          # indicate which worker to use
          worker: fork_worker
          fork_worker:
            # set this or num_workers; num_workers takes precedence
            num_workers_per_core: 1
            # how often to check if there is more work
            poll_interval: 1
            # how often to start more workers
            spawn_interval: 0.01
            # how often to record our existence
            heartbeat_interval: 15
            # minimum time a working worker will live
            child_lifetime: 10
            # kill off jobs this long before their deadlines
            stop_jobs_early: 15

    This spawns child processes to do work.  Each child process does at
    most one work unit.  If `num_workers` is set, at most this many
    concurrent workers will be running at a time.  If `num_workers` is
    not set but `num_workers_per_core` is, the maximum number of workers
    is a multiple of the number of processor cores available.  The
    default setting is 1 worker per core, but setting this higher can
    be beneficial if jobs are alternately network- and CPU-bound.

    The parent worker runs a fairly simple state machine.  It awakens
    on startup, whenever a child process exits, or after a timeout.
    When it awakens, it checks on the status of all of its children,
    and collects the exit status of those that have finished.  If any
    failed or reported no more work, the timeout is set to
    `poll_interval`, and no more workers are started until that
    timeout has passed.  Otherwise, if it is not running the maximum
    number of workers, it starts one exactly and sets the timeout to
    `spawn_interval`.

    This means that if the system is operating normally, and there is
    work to do, then it will start all of its workers in `num_workers`
    times `spawn_interval` time.  If `spawn_interval` is 0, then any
    time the system thinks it may have work to do, it will spawn the
    maximum number of processes immediately, each of which will
    connect to Redis.  If the system runs out of work, or if it starts
    all of its workers, it will check for work or system shutdown
    every `poll_interval`.  The parent worker will contact Redis,
    recording its state and retrieving the global mode, every
    `heartbeat_interval`.

    Every `heartbeat_interval` the parent also checks on the jobs its
    children are running.  If any of them are overdue now or being
    worked on by other workers, the parent will kill them to avoid
    having multiple workers doing the same work unit.  Furthermore, if
    any childrens' jobs will expire within `stop_jobs_early` seconds,
    those jobs will be killed too even if they aren't expired yet, and
    any jobs killed this way will be marked failed if they are still
    owned by the same child worker.  If `stop_jobs_early` is at least
    `heartbeat_interval`, this will reliably cause jobs that take
    longer than the expiry interval (default 300 seconds) to be killed
    off rather than retried.

    '''

    '''Several implementation notes:

    fork() and logging don't mix.  While we're better off here than
    using :mod:`multiprocessing` (which forks from a thread, so the
    child can start up with a log handler locked) there are still
    several cases like syslog or dblogger handlers that depend on a
    network connection that won't be there in the child.

    That means that *nothing in the parent gets to log at all*.  We
    need to maintain a separate child process to do logging.

    ``debug_worker`` is a hidden additional configuration option.  It
    can cause a lot of boring information to get written to the log.
    It is a list of keywords.  ``children`` logs all process creation
    and destruction.  ``loop`` shows what we're thinking in the main
    loop.

    '''

    config_name = 'fork_worker'
    default_config = {
        'num_workers': None,
        'num_workers_per_core': 1,
        'poll_interval': 1,
        'spawn_interval': 0.01,
        'heartbeat_interval': 15,
        'child_lifetime': 10,
        'stop_jobs_early': 15,
        'debug_worker': None,
    }

    @classmethod
    def config_get(cls, config, key):
        return config.get(key, cls.default_config[key])

    def __init__(self, config):
        '''Create a new forking worker.

        This starts up with no children and no logger, and it will
        set itself up and contact Redis as soon as :meth:`run` is called.

        :param dict config: ``rejester`` config dictionary

        '''
        super(ForkWorker, self).__init__(config)
        c = config.get(self.config_name, {})
        self.num_workers = self.config_get(c, 'num_workers')
        if not self.num_workers:
            num_cores = 0
            # This replicates the Good Unix case of multiprocessing.cpu_count()
            try:
                num_cores = os.sysconf('SC_NPROCESSORS_ONLN')
            except (ValueError, OSError, AttributeError):
                pass
            num_cores = max(1, num_cores)
            num_workers_per_core = self.config_get(c, 'num_workers_per_core')
            self.num_workers = num_workers_per_core * num_cores
        self.poll_interval = self.config_get(c, 'poll_interval')
        self.spawn_interval = self.config_get(c, 'spawn_interval')
        self.heartbeat_interval = self.config_get(c, 'heartbeat_interval')
        self.heartbeat_deadline = time.time() - 1  # due now
        self.child_lifetime = self.config_get(c, 'child_lifetime')
        self.stop_jobs_early = self.config_get(c, 'stop_jobs_early')
        self.debug_worker = c.get('debug_worker', [])
        self.children = set()
        self.log_child = None
        self.log_fd = None
        self.shutting_down = False
        self.last_mode = None
        self.old_sigabrt = None
        self.old_sigint = None
        self.old_sigpipe = None
        self.old_sigterm = None

    @staticmethod
    def pid_is_alive(pid):
        try:
            os.kill(pid, 0)
            # This doesn't actually send a signal, but does still do the
            # "is pid alive?" check.  If we didn't get an exception, it is.
            return True
        except OSError, e:
            if e.errno == errno.ESRCH:
                # "No such process"
                return False
            raise

    def set_signal_handlers(self):
        '''Set some signal handlers.

        These react reasonably to shutdown requests, and keep the
        logging child alive.

        '''
        def handler(f):
            def wrapper(signum, backtrace):
                return f()
            return wrapper

        self.old_sigabrt = signal.signal(signal.SIGABRT,
                                         handler(self.scram))
        self.old_sigint = signal.signal(signal.SIGINT,
                                        handler(self.stop_gracefully))
        self.old_sigpipe = signal.signal(signal.SIGPIPE,
                                         handler(self.live_log_child))
        signal.siginterrupt(signal.SIGPIPE, False)
        self.old_sigterm = signal.signal(signal.SIGTERM,
                                         handler(self.stop_gracefully))

    def clear_signal_handlers(self):
        '''Undo :meth:`set_signal_handlers`.

        Not only must this be done on shutdown, but after every fork
        call too.

        '''
        signal.signal(signal.SIGABRT, self.old_sigabrt)
        signal.signal(signal.SIGINT, self.old_sigint)
        signal.signal(signal.SIGPIPE, self.old_sigpipe)
        signal.signal(signal.SIGTERM, self.old_sigterm)

    def log(self, level, message):
        '''Write a log message via the child process.

        The child process must already exist; call :meth:`live_log_child`
        to make sure.  If it has died in a way we don't expect then
        this will raise :const:`signal.SIGPIPE`.

        '''
        if self.log_fd is not None:
            prefix = struct.pack('ii', level, len(message))
            os.write(self.log_fd, prefix)
            os.write(self.log_fd, message)

    def debug(self, group, message):
        '''Maybe write a debug-level log message.

        In particular, this gets written if the hidden `debug_worker`
        option contains `group`.

        '''
        if group in self.debug_worker:
            if 'stdout' in self.debug_worker:
                print message
            self.log(logging.DEBUG, message)

    def log_spewer(self, gconfig, fd):
        '''Child process to manage logging.

        This reads pairs of lines from `fd`, which are alternating
        priority (Python integer) and message (unformatted string).

        '''
        setproctitle('rejester fork_worker log task')
        yakonfig.set_default_config([yakonfig, dblogger], config=gconfig)
        try:
            while True:
                prefix = os.read(fd, struct.calcsize('ii'))
                level, msglen = struct.unpack('ii', prefix)
                msg = os.read(fd, msglen)
                logger.log(level, msg)
        except Exception, e:
            logger.critical('log writer failed', exc_info=e)
            raise

    def start_log_child(self):
        '''Start the logging child process.'''
        self.stop_log_child()
        gconfig = yakonfig.get_global_config()
        read_end, write_end = os.pipe()
        pid = os.fork()
        if pid == 0:
            # We are the child
            self.clear_signal_handlers()
            os.close(write_end)
            yakonfig.clear_global_config()
            self.log_spewer(gconfig, read_end)
            sys.exit(0)
        else:
            # We are the parent
            self.debug('children', 'new log child with pid {}'.format(pid))
            self.log_child = pid
            os.close(read_end)
            self.log_fd = write_end

    def stop_log_child(self):
        '''Stop the logging child process.'''
        if self.log_fd:
            os.close(self.log_fd)
            self.log_fd = None
        if self.log_child:
            try:
                self.debug('children', 'stopping log child with pid {}'
                           .format(self.log_child))
                os.kill(self.log_child, signal.SIGTERM)
                os.waitpid(self.log_child, 0)
            except OSError, e:
                if e.errno == errno.ESRCH or e.errno == errno.ECHILD:
                    # already gone
                    pass
                else:
                    raise
            self.log_child = None

    def live_log_child(self):
        '''Start the logging child process if it died.'''
        if not (self.log_child and self.pid_is_alive(self.log_child)):
            self.start_log_child()

    def do_some_work(self, can_start_more):
        '''Run one cycle of the main loop.

        If the log child has died, restart it.  If any of the worker
        children have died, collect their status codes and remove them
        from the child set.  If there is a worker slot available, start
        exactly one child.

        :param bool can_start_more: Allowed to start a child?
        :return:  Time to wait before calling this function again

        '''
        any_happy_children = False
        any_sad_children = False
        any_bored_children = False

        self.debug('loop', 'starting work loop, can_start_more={!r}'
                   .format(can_start_more))

        # See if anyone has died
        while True:
            try:
                pid, status = os.waitpid(-1, os.WNOHANG)
            except OSError, e:
                if e.errno == errno.ECHILD:
                    # No children at all
                    pid = 0
                else:
                    raise
            if pid == 0:
                break
            elif pid == self.log_child:
                self.debug('children',
                           'log child with pid {} exited'.format(pid))
                self.start_log_child()
            elif pid in self.children:
                self.children.remove(pid)
                if os.WIFEXITED(status):
                    code = os.WEXITSTATUS(status)
                    self.debug('children',
                               'worker {} exited with code {}'
                               .format(pid, code))
                    if code == SingleWorker.EXIT_SUCCESS:
                        any_happy_children = True
                    elif code == SingleWorker.EXIT_EXCEPTION:
                        self.log(logging.WARNING,
                                 'child {} reported failure'.format(pid))
                        any_sad_children = True
                    elif code == SingleWorker.EXIT_BORED:
                        any_bored_children = True
                    else:
                        self.log(logging.WARNING,
                                 'child {} had odd exit code {}'
                                 .format(pid, code))
                elif os.WIFSIGNALED(status):
                    self.log(logging.WARNING,
                             'child {} exited with signal {}'
                             .format(pid, os.WTERMSIG(status)))
                    any_sad_children = True
                else:
                    self.log(logging.WARNING,
                             'child {} went away with unknown status {}'
                             .format(pid, status))
                    any_sad_children = True
            else:
                self.log(logging.WARNING,
                         'child {} exited, but we don\'t recognize it'
                         .format(pid))

        # ...what next?
        # (Don't log anything here; either we logged a WARNING message
        # above when things went badly, or we're in a very normal flow
        # and don't want to spam the log)
        if any_sad_children:
            self.debug('loop', 'exit work loop with sad child')
            return self.poll_interval

        if any_bored_children:
            self.debug('loop', 'exit work loop with no work')
            return self.poll_interval

        # This means we get to start a child, maybe.
        if can_start_more and len(self.children) < self.num_workers:
            pid = os.fork()
            if pid == 0:
                # We are the child
                self.clear_signal_handlers()
                if self.log_fd:
                    os.close(self.log_fd)
                LoopWorker.as_child(yakonfig.get_global_config(),
                                    parent=self.worker_id)
                # This should never return, but just in case
                sys.exit(SingleWorker.EXIT_EXCEPTION)
            else:
                # We are the parent
                self.debug('children', 'new worker with pid {}'.format(pid))
                self.children.add(pid)
                self.debug('loop', 'exit work loop with a new worker')
                return self.spawn_interval

        # Absolutely nothing is happening; which means we have all
        # of our potential workers and they're doing work
        self.debug('loop', 'exit work loop with full system')
        return self.poll_interval

    def check_spinning_children(self):
        '''Stop children that are working on overdue jobs.'''
        child_jobs = self.task_master.get_child_work_units(self.worker_id)
        # We will kill off any jobs that are due before "now".  This
        # isn't really now now, but now plus a grace period to make
        # sure spinning jobs don't get retried.
        now = time.time() + self.stop_jobs_early
        for child, wu in child_jobs.iteritems():
            if wu.worker_id == child and wu.expires > now:
                # We are still working on a not-overdue job
                continue
            # So either someone else is doing its work or it's just overdue
            environment = self.task_master.get_heartbeat(child)
            if 'pid' not in environment:
                continue  # derp
            if environment['pid'] not in self.children:
                continue  # derp
            os.kill(environment['pid'], signal.SIGTERM)
            # This will cause the child to die, and do_some_work will
            # reap it; but we'd also like the job to fail if possible
            if wu.worker_id == child:
                wu.data['traceback'] = 'job expired'
                wu.fail()

    def stop_gracefully(self):
        '''Refuse to start more processes.

        This runs in response to SIGINT or SIGTERM; if this isn't a
        background process, control-C and a normal ``kill`` command
        cause this.

        '''
        if self.shutting_down:
            self.log(logging.INFO,
                     'second shutdown request, shutting down now')
            self.scram()
        else:
            self.log(logging.INFO, 'shutting down after current jobs finish')
            self.shutting_down = True

    def stop_all_children(self):
        '''Kill all workers.'''
        # There's an unfortunate race condition if we try to log this
        # case: we can't depend on the logging child actually receiving
        # the log message before we kill it off.  C'est la vie...
        self.stop_log_child()
        for pid in self.children:
            try:
                os.kill(pid, signal.SIGTERM)
                os.waitpid(pid, 0)
            except OSError, e:
                if e.errno == errno.ESRCH or e.errno == errno.ECHILD:
                    # No such process
                    pass
                else:
                    raise

    def scram(self):
        '''Kill all workers and die ourselves.

        This runs in response to SIGABRT, from a specific invocation
        of the ``kill`` command.  It also runs if
        :meth:`stop_gracefully` is called more than once.

        '''
        self.stop_all_children()
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        sys.exit(2)

    def run(self):
        '''Run the main loop.

        This is fairly invasive: it sets a bunch of signal handlers
        and spawns off a bunch of child processes.

        '''
        setproctitle('rejester fork_worker for namespace {}'
                     .format(self.config.get('namespace', None)))
        self.set_signal_handlers()
        try:
            self.start_log_child()
            while True:
                can_start_more = not self.shutting_down
                if time.time() >= self.heartbeat_deadline:
                    mode = self.heartbeat()
                    if mode != self.last_mode:
                        self.log(logging.INFO,
                                 'rejester global mode is {!r}'.format(mode))
                        self.last_mode = mode
                    self.heartbeat_deadline = (time.time() +
                                               self.heartbeat_interval)
                    self.check_spinning_children()
                else:
                    mode = self.last_mode
                if mode != self.task_master.RUN:
                    can_start_more = False
                interval = self.do_some_work(can_start_more)
                # Normal shutdown case
                if len(self.children) == 0:
                    if mode == self.task_master.TERMINATE:
                        self.log(logging.INFO,
                                 'stopping for rejester global shutdown')
                        break
                    if self.shutting_down:
                        self.log(logging.INFO,
                                 'stopping in response to signal')
                        break
                time.sleep(interval)
        except Exception:
            self.log(logging.CRITICAL,
                     'uncaught exception in worker: ' + traceback.format_exc())
        finally:
            # See the note in run_workers() above.  clear_signal_handlers()
            # calls signal.signal() which explicitly affects the current
            # process, parent or child.
            self.clear_signal_handlers()



'''
http://github.com/diffeo/rejester

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import time
import uuid
import json
import logging
import os
import random
import socket
import struct
import atexit
import contextlib
from uuid import UUID
from functools import wraps
from collections import defaultdict
from operator import mul

import redis

from rejester.exceptions import EnvironmentError, LockError, \
    ProgrammerError
from rejester._redis import RedisBase

logger = logging.getLogger(__name__)

def nice_identifier():
    'do not use uuid.uuid4, because it can block'
    big = reduce(mul, struct.unpack('<LLLL', os.urandom(16)), 1)
    big = big % 2**128
    return uuid.UUID(int=big).hex

class Registry(RedisBase):
    '''Store string-keyed dictionaries in Redis.

    Provides a centralized storage mechanism for dictionaries,
    including atomic operations for moving (key, value) pairs between
    dictionaries, and incrementing counts.

    Many operations on the registry require getting a lock via
    the database, for instance

    >>> with registry.lock() as session:
    ...   value = session.get(k1, k2)

    The lock mechanism ensures that no two Registry objects do work
    concurrently, even running on separate systems.  Specific method
    descriptions will note if they can run without a lock.  In
    general, read-only operations will always run successfully without
    a lock but will check for the correct lock if one is given;
    certain very simple operations do no lock checking at all; and
    read-write operations always require a lock.

    The basic data object is a string-keyed dictionary stored under
    some key.  The dictionary keys are also stored in a prioritized
    list.  This in effect provides two levels of dictionary, using the
    Redis key and the dictionary key.  The registry makes an effort to
    store all types of object as values, potentially serializing them
    into JSON.

    '''

    # This makes extensive use of Lua scripts in Redis; see
    # http://redis.io/commands/eval for basics.  This provides atomicity
    # (Redis never runs anything in parallel with a Lua script), but
    # also means that much of the logic is in a hard-to-watch script,
    # and that performance of these scripts is pretty critical.

    def __init__(self, config):
        '''
        Initialize the registry using a config
        '''
        super(Registry, self).__init__(config)

        ## populated only when lock is acquired
        self._session_lock_identifier = None

    def _conn(self):
        '''debugging aid to easily grab a connection from the connection pool

        '''
        conn = redis.Redis(connection_pool=self.pool)
        return conn

    @property
    def _lock_name(self):
        '''Name of the key for the global lock.'''
        return self._namespace('global_registry_lock')

    def _acquire_lock(self, identifier, atime=30, ltime=5):
        '''Acquire a lock for a given identifier.

        If the lock cannot be obtained immediately, keep trying at random
        intervals, up to 3 seconds, until `atime` has passed.  Once the
        lock has been obtained, continue to hold it for `ltime`.

        :param str identifier: lock token to write
        :param int atime: maximum time (in seconds) to acquire lock
        :param int ltime: maximum time (in seconds) to own lock
        :return: `identifier` if the lock was obtained, :const:`False`
          otherwise

        '''
        conn = redis.Redis(connection_pool=self.pool)
        end = time.time() + atime
        while end > time.time():
            if conn.set(self._lock_name, identifier, ex=ltime, nx=True):
                # logger.debug("won lock %s" % self._lock_name)
                return identifier
            sleep_time = random.uniform(0, 3)
            time.sleep(sleep_time)

        logger.warn('failed to acquire lock %s for %f seconds',
                    self._lock_name, atime)
        return False


    def re_acquire_lock(self, ltime=5):
        '''Re-acquire the lock.

        You must already own the lock; this is best called from
        within a :meth:`lock` block.

        :param int ltime: maximum time (in seconds) to own lock
        :return: the session lock identifier
        :raise rejester.exceptions.EnvironmentError:
          if we didn't already own the lock

        '''
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
          return redis.call("expire", KEYS[1], ARGV[2])
        else
          return -1
        end
        ''')
        ret = script(keys=[self._lock_name],
                     args=[self._session_lock_identifier, ltime])
        if ret != 1:
            raise EnvironmentError('failed to re-acquire lock')

        # logger.debug('re-acquired lock %s', self._lock_name)
        return self._session_lock_identifier

    def _release_lock(self, identifier):
        '''Release the lock.

        This requires you to actually have owned the lock.  On return
        you definitely do not own it, but if somebody else owned it
        before calling this function, they still do.

        :param str identifier: the session lock identifier
        :return: :const:`True` if you actually did own the lock,
          :const:`False` if you didn't

        '''
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            return redis.call("del", KEYS[1])
        else
            return -1
        end
        ''')
        num_keys_deleted = script(keys=[self._lock_name],
                                  args=[identifier])
        return (num_keys_deleted == 1)

    @contextlib.contextmanager
    def lock(self, atime=30, ltime=5, identifier=None):
        '''Context manager to acquire the namespace global lock.

        This is typically used for multi-step registry operations,
        such as a read-modify-write sequence::

            with registry.lock() as session:
                d = session.get('dict', 'key')
                del d['traceback']
                session.set('dict', 'key', d)

        Callers may provide their own `identifier`; if they do, they
        must ensure that it is reasonably unique (e.g., a UUID).
        Using a stored worker ID that is traceable back to the lock
        holder is a good practice.

        :param int atime: maximum time (in seconds) to acquire lock
        :param int ltime: maximum time (in seconds) to own lock
        :param str identifier: worker-unique identifier for the lock

        '''
        if identifier is None:
            identifier = nice_identifier()
        if self._acquire_lock(identifier, atime, ltime) != identifier:
            raise LockError("could not acquire lock")
        try:
            self._session_lock_identifier = identifier
            yield self
        finally:
            self._release_lock(identifier)
            self._session_lock_identifier = None

    def read_lock(self):
        '''Find out who currently owns the namespace global lock.

        This is purely a diagnostic tool.  If you are trying to get
        the global lock, it is better to just call :meth:`lock`, which
        will atomically get the lock if possible and retry.

        :return: session identifier of the lock holder, or :const:`None`

        '''
        return redis.Redis(connection_pool=self.pool).get(self._lock_name)

    def force_clear_lock(self):
        '''Kick out whoever currently owns the namespace global lock.

        This is intended as purely a last-resort tool.  If another
        process has managed to get the global lock for a very long time,
        or if it requested the lock with a long expiration and then
        crashed, this can make the system functional again.  If the
        original lock holder is still alive, its session calls may fail
        with exceptions.

        '''
        return redis.Redis(connection_pool=self.pool).delete(self._lock_name)

    def _encode(self, data):
        '''Redis hash's store strings in the keys and values.  Since we want
        to store UUIDs and tuples of UUIDs, we cannot just use json.
        Instead, we employ a very simply decoding to turn various
        strings into datatypes we understand.
        '''
        if isinstance(data, tuple):
            return 't:' + '-'.join([self._encode(item) for item in data])
        elif isinstance(data, UUID):
            return 'u:' + str(data.int)
        else:
            try:
                return 'j:' + json.dumps(data)
            except Exception, exc:
                logger.error('Fail to encode data %r', data, exc_info=True)
                raise TypeError

    def _decode(self, string):
        '''Redis hash's store strings in the keys and values.  Since we want
        to store UUIDs and tuples of UUIDs, we cannot just use json.
        Instead, we employ a very simply decoding to turn various
        strings into datatypes we understand.

        Note, that redis would convert things into a string
        representation that we could 'eval' here, but it is not a good
        practice from a security standpoint to eval data from a
        database.
        '''
        if len(string) == 0:
            return None
        if string[0] == 't':
            return tuple([self._decode(item) for item in
                          string[2:].split('-')])
        elif string[0] == 'u':
            return UUID(int=int(string[2:]))
        elif string[0] == 'j':
            try:
                if string[2:]:
                    return json.loads(string[2:])
                else:
                    None
            except ValueError, exc:
                logger.critical('%r --> tried to json.loads(%r)', string, string[2:], exc_info=True)
                raise
        else:
            try:
                return float(string)
            except:
                pass
            ## Raise a type error on all other types of data
            logger.error('Fail to decode string %r', string, exc_info=True)
            raise TypeError

    def update(self, dict_name, mapping=None, priorities=None, expire=None,
               locks=None):
        '''Add mapping to a dictionary, replacing previous values

        Can be called with only dict_name and expire to refresh the
        expiration time.

        NB: locks are only enforced if present, so nothing prevents
        another caller from coming in an modifying data without using
        locks.

        :param mapping: a dict of keys and values to update in
          dict_name.  Must be specified if priorities is specified.
        :param priorities: a dict with the same keys as those in
          mapping that provides a numerical value indicating the
          priority to assign to that key.  Default sets 0 for all keys.
        :param int expire: if specified, then dict_name will be set to
          expire in that many seconds.
        :param locks: a dict with the same keys as those in the
          mapping.  Before making any particular update, this function
          checks if a key is present in a 'locks' table for this dict,
          and if so, then its value must match the value provided in the
          input locks dict for that key.  If not, then the value
          provided in the locks dict is inserted into the 'locks' table.
          If the locks parameter is None, then no lock checking is
          performed.
        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        if priorities is None:
            ## set all priorities to zero
            priorities = defaultdict(int)
        if locks is None:
            ## set all locks to None
            locks = defaultdict(lambda: '')
        if not (expire is None or isinstance(expire, int)):
            raise ProgrammerError('expire must be int or unspecified')
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            for i = 3, #ARGV, 4  do
                if ARGV[i+3] ~= 'j:""' then
                    local curr_lock = redis.call("hget",  KEYS[4], ARGV[i])
                    if curr_lock and curr_lock ~= ARGV[i+3] then
                        return {-1, ARGV[i], curr_lock, ARGV[i+3]}
                    end
                    redis.call("hset",  KEYS[4], ARGV[i], ARGV[i+3])
                end
            end
            for i = 3, #ARGV, 4  do
                redis.call("hset",  KEYS[2], ARGV[i], ARGV[i+1])
                redis.call("zadd",  KEYS[3], ARGV[i+2], ARGV[i])
            end
            if tonumber(ARGV[2]) ~= nil then
                redis.call("expire", KEYS[2], ARGV[2])
                redis.call("expire", KEYS[3], ARGV[2])
            end
            return {1, 0}
        else
            -- ERROR: No longer own the lock
            return {0, 0}
        end
        ''')
        dict_name = self._namespace(dict_name)
        if mapping is None:
            mapping = {}
        items = []
        ## This flattens the dictionary into a list
        for key, value in mapping.iteritems():
            items.append(self._encode(key))
            items.append(self._encode(value))
            items.append(priorities[key])
            items.append(self._encode(locks[key]))

        #logger.debug('update %r %r', dict_name, items)
        res = script(keys=[self._lock_name,
                           dict_name,
                           dict_name + 'keys',
                           dict_name + '_locks'],
                     args=[self._session_lock_identifier, expire] + items)
        if res[0] == 0:
            raise EnvironmentError(
                'Unable to add items to %s in registry' % dict_name)
        elif res[0] == -1:
            raise EnvironmentError(
                'lost lock on key=%r owned by %r not %r in %s' 
                % (self._decode(res[1]), res[2], res[3], dict_name))


    def reset_priorities(self, dict_name, priority):
        '''set all priorities in dict_name to priority

        :type priority: float or int
        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        ## see comment above for script in update
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            local keys = redis.call('ZRANGE', KEYS[2], 0, -1)
            for i, next_key in ipairs(keys) do
                redis.call("zadd",  KEYS[2], ARGV[2], next_key)
            end
            return 1
        else
            -- ERROR: No longer own the lock
            return 0
        end
        ''')
        dict_name = self._namespace(dict_name)
        res = script(keys=[self._lock_name,
                           dict_name + 'keys'],
                     args=[self._session_lock_identifier,
                           priority])
        if not res:
            # We either lost the lock or something else went wrong
            raise EnvironmentError(
                'Unable to add items to %s in registry' % dict_name)


    def popmany(self, dict_name, *keys):
        '''Remove one or more keys from a dictionary.

        If any of the `keys` are not present, they are silently ignored.
        
        The actual deletion operation is atomic and does not require a
        session lock, but nothing stops another operation from creating
        the deleted keys immediately afterwards.  You may call this with
        or without a session lock, but the operation will fail if some
        other worker holds one.

        :param str dict_name: name of dictionary to modify
        :param str keys: names of keys to remove
        :returns: number of keys removed
        :raises rejester.exceptions.LockError: if the session lock timed
          out, or if this was called without a session lock and some other
          worker holds one
        '''
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        local lock_holder = redis.call("get", KEYS[1])
        if lock_holder == ARGV[1]
        then
            local count = 0
            for i = 2, #ARGV do
                if (redis.call("hexists", KEYS[2], ARGV[i]) == 1)
                then
                    count = count + 1
                    redis.call("hdel", KEYS[2], ARGV[i])
                    redis.call("zrem", KEYS[3], ARGV[i])
                end
            end
            return count
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')
        res = script(keys=[self._lock_name,
                           self._namespace(dict_name),
                           self._namespace(dict_name) + "keys"],
                     args=([self._session_lock_identifier] +
                           [self._encode(key) for key in keys]))
        if res < 0:
            raise LockError()
        return res

    def len(self, dict_name, priority_min='-inf', priority_max='+inf'):
        '''Get the number of items in (part of) a dictionary.

        Returns number of items in `dict_name` within
        [`priority_min`, `priority_max`].  This is similar to
        ``len(filter(dict_name, priority_min, priority_max))`` but
        does not actually retrieve the items.

        This is a read-only operation that does not require or
        honor a session lock.

        :param str dict_name: dictionary name to query
        :param float priority_min: lowest priority score
        :param float priority_max: highest priority score
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        return conn.zcount(dict_name + 'keys', priority_min, priority_max)

    def getitem_reset(
            self, dict_name, priority_min='-inf', priority_max='+inf',
            new_priority=0,
            lock=None
        ):
        '''Select an item and update its priority score.

        The item comes from `dict_name`, and has the lowest score at
        least `priority_min` and at most `priority_max`.  If some item
        is found, change its score to `new_priority` and return it.

        This runs as a single atomic operation but still requires a
        session lock.

        :param str dict_name: source dictionary
        :param float priority_min: lowest score
        :param float priority_max: highest score
        :param float new_priority: new score
        :param str lock: lock value for the item
        :return: pair of (key, value) if an item was reprioritized, or
          :const:`None`

        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- remove next item of dict_name
            local next_key, next_priority = redis.call("zrangebyscore",
                KEYS[3], ARGV[2], ARGV[3], "WITHSCORES", "LIMIT", 0, 1)[1]

            if not next_key then
                return {}
            end

            local next_val = redis.call("hget", KEYS[2], next_key)
            redis.call("zadd", KEYS[3], ARGV[4], next_key)
            if ARGV[5] then
                redis.call("hset", KEYS[4], next_key, ARGV[5])
                redis.call("hset", KEYS[4], ARGV[5], next_key)
            else
                local old_key2 = redis.call("hget", KEYS[4], next_key)
                redis.call("hdel", KEYS[4], next_key)
                redis.call("hdel", KEYS[4], old_key2)
            end

            return {next_key, next_val}
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')
        dict_name = self._namespace(dict_name)
        key_value = script(keys=[self._lock_name,
                                 dict_name,
                                 dict_name + 'keys',
                                 dict_name + '_locks'],
                           args=[self._session_lock_identifier,
                                 priority_min, priority_max,
                                 new_priority,
                                 self._encode(lock)])
        if key_value == -1:
            raise KeyError(
                'Registry failed to return an item from %s' % dict_name)

        if key_value == []:
            return None

        return self._decode(key_value[0]), self._decode(key_value[1])

    def popitem(self, dict_name, priority_min='-inf', priority_max='+inf'):
        '''Select an item and remove it.

        The item comes from `dict_name`, and has the lowest score
        at least `priority_min` and at most `priority_max`.  If some
        item is found, remove it from `dict_name` and return it.

        This runs as a single atomic operation but still requires a
        session lock.

        :param str dict_name: source dictionary
        :param float priority_min: lowest score
        :param float priority_max: highest score
        :return: pair of (key, value) if an item was popped, or
          :const:`None`

        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- remove next item of dict_name
            local next_key, next_priority = redis.call("zrangebyscore",
                KEYS[3], ARGV[2], ARGV[3], "WITHSCORES", "LIMIT", 0, 1)[1]

            if not next_key then
                return {}
            end
            
            redis.call("zrem", KEYS[3], next_key)
            local next_val = redis.call("hget", KEYS[2], next_key)
            -- zrem removed it from list, so also remove from hash
            redis.call("hdel", KEYS[2], next_key)
            return {next_key, next_val}
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')
        dict_name = self._namespace(dict_name)
        key_value = script(keys=[self._lock_name,
                                 dict_name,
                                 dict_name + "keys"],
                           args=[self._session_lock_identifier,
                                 priority_min,
                                 priority_max])
        if key_value == -1:
            raise KeyError(
                'Registry failed to return an item from %s' % dict_name)

        if key_value == []:
            return None

        return self._decode(key_value[0]), self._decode(key_value[1])

    def popitem_move(self, from_dict, to_dict,
                     priority_min='-inf', priority_max='+inf'):
        '''Select an item and move it to another dictionary.

        The item comes from `from_dict`, and has the lowest score
        at least `priority_min` and at most `priority_max`.  If some
        item is found, remove it from `from_dict`, add it to `to_dict`,
        and return it.

        This runs as a single atomic operation but still requires a
        session lock.

        :param str from_dict: source dictionary
        :param str to_dict: destination dictionary
        :param float priority_min: lowest score
        :param float priority_max: highest score
        :return: pair of (key, value) if an item was moved, or
          :const:`None`

        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- find the next key and priority
            local next_items = redis.call("zrangebyscore", KEYS[3],
                ARGV[2], ARGV[3], "WITHSCORES", "LIMIT", 0, 1)
            local next_key = next_items[1]
            local next_priority = next_items[2]
            
            if not next_key then
                return {}
            end

            -- remove next item of from_dict
            redis.call("zrem", KEYS[3], next_key)

            local next_val = redis.call("hget", KEYS[2], next_key)
            -- zrem removed it from list, so also remove from hash
            redis.call("hdel", KEYS[2], next_key)

            -- put it in to_dict
            redis.call("hset", KEYS[4], next_key, next_val)
            redis.call("zadd", KEYS[5], next_priority, next_key)

            return {next_key, next_val, next_priority}
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')
        key_value = script(keys=[self._lock_name, 
                                 self._namespace(from_dict),
                                 self._namespace(from_dict) + 'keys',
                                 self._namespace(to_dict),
                                 self._namespace(to_dict) + 'keys'],
                           args=[self._session_lock_identifier,
                                 priority_min, priority_max])

        if key_value == []:
            return None

        if None in key_value or key_value == -1:
            raise KeyError(
                'Registry.popitem_move(%r, %r) --> %r' %
                (from_dict, to_dict, key_value))

        return self._decode(key_value[0]), self._decode(key_value[1])

    def move(self, from_dict, to_dict, mapping, priority=None):
        '''Move keys between dictionaries, possibly with changes.

        Every key in `mapping` is removed from `from_dict`, and added to
        `to_dict` with its corresponding value.  The priority will be
        `priority`, if specified, or else its current priority.

        This operation on its own is atomic and does not require a
        session lock; however, it does require you to pass in the
        values, which probably came from a previous query call.  If
        you do not call this with a session lock but some other caller
        has one, you will get :class:`rejester.LockError`.  If you do
        have a session lock, this will check that you still have it.

        :param str from_dict: name of original dictionary
        :param str to_dict: name of target dictionary
        :param dict mapping: keys to move with new values
        :param int priority: target priority, or :const:`None` to use existing
        :raise rejester.LockError: if the session lock timed out
        :raise rejester.EnvironmentError: if some items didn't move

        '''
        items = []
        ## This flattens the dictionary into a list
        for key, value in mapping.iteritems():
            key = self._encode(key)
            items.append(key)
            value = self._encode(value)
            items.append(value)

        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        local lock_holder = redis.call("get", KEYS[1])
        if (not lock_holder) or (lock_holder == ARGV[1])
        then
            local count = 0
            for i = 3, #ARGV, 2  do
                -- find the priority of the item
                local next_priority = ARGV[2]
                if next_priority == "" then
                    next_priority = redis.call("zscore", KEYS[3], ARGV[i])
                end
                -- remove item from the sorted set and the real data
                redis.call("hdel", KEYS[2], ARGV[i])
                redis.call("zrem", KEYS[3], ARGV[i])
                -- put it in to_dict
                redis.call("hset", KEYS[4], ARGV[i], ARGV[i+1])
                redis.call("zadd", KEYS[5], next_priority, ARGV[i])
                count = count + 1
            end
            return count
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')

        # Experimental evidence suggests that redis-py passes
        # *every* value as a string, maybe unless it's obviously
        # a number.  Empty string is an easy "odd" value to test for.
        if priority is None: priority = ''
        num_moved = script(keys=[self._lock_name,
                                 self._namespace(from_dict),
                                 self._namespace(from_dict) + "keys",
                                 self._namespace(to_dict),
                                 self._namespace(to_dict) + "keys"],
                           args=[self._session_lock_identifier,
                                 priority] + items)
        if num_moved == -1:
            raise LockError()
        if num_moved != len(items) / 2:
            raise EnvironmentError(
                'Registry failed to move all: num_moved = %d != %d len(items)'
                % (num_moved, len(items)))


    def move_all(self, from_dict, to_dict):
        '''Move everything from one dictionary to another.

        This can be expensive if the source dictionary is large.

        This always requires a session lock.

        :param str from_dict: source dictionary
        :param str to_dict: destination dictionary

        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            local count = 0
            local keys = redis.call('ZRANGE', KEYS[3], 0, -1)
            for i, next_key in ipairs(keys) do
                -- get the value and priority for this key
                local next_val = redis.call("hget", KEYS[2], next_key)
                local next_priority = redis.call("zscore", KEYS[3], next_key)
                -- remove item of from_dict
                redis.call("zrem", KEYS[3], next_key)
                -- also remove from hash
                redis.call("hdel", KEYS[2], next_key)
                -- put it in to_dict
                redis.call("hset",  KEYS[4], next_key, next_val)
                redis.call("zadd", KEYS[5], next_priority, next_key)
                count = count + 1
            end
            return count
        else
            -- ERROR: No longer own the lock
            return 0
        end
        ''')
        num_moved = script(keys=[self._lock_name,
                                 self._namespace(from_dict), 
                                 self._namespace(from_dict) + 'keys',
                                 self._namespace(to_dict),
                                 self._namespace(to_dict) + 'keys'],
                           args=[self._session_lock_identifier])
        return num_moved


    def pull(self, dict_name):
        '''Get the entire contents of a single dictionary.

        This operates without a session lock, but is still atomic.  In
        particular this will run even if someone else holds a session
        lock and you do not.

        This is only suitable for "small" dictionaries; if you have
        hundreds of thousands of items or more, consider
        :meth:`filter` instead to get a subset of a dictionary.

        :param str dict_name: name of the dictionary to retrieve
        :return: corresponding Python dictionary

        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        res = conn.hgetall(dict_name)
        split_res = {self._decode(key):
                     self._decode(value)
                     for key, value in res.iteritems()}
        return split_res

    def filter(self, dict_name, priority_min='-inf', priority_max='+inf',
               start=0, limit=None):
        '''Get a subset of a dictionary.

        This retrieves only keys with priority scores greater than or
        equal to `priority_min` and less than or equal to `priority_max`.
        Of those keys, it skips the first `start` ones, and then returns
        at most `limit` keys.

        With default parameters, this retrieves the entire dictionary,
        making it a more expensive version of :meth:`pull`.  This can
        be used to limit the dictionary by priority score, for instance
        using the score as a time stamp and only retrieving values
        before or after a specific time; or it can be used to get
        slices of the dictionary if there are too many items to use
        :meth:`pull`.

        This is a read-only operation and does not require a session
        lock, but if this is run in a session context, the lock will
        be honored.

        :param str dict_name: name of the dictionary to retrieve
        :param float priority_min: lowest score to retrieve
        :param float priority_max: highest score to retrieve
        :param int start: number of items to skip
        :param int limit: number of items to retrieve
        :return: corresponding (partial) Python dictionary
        :raise rejester.LockError: if the session lock timed out

        '''
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if (ARGV[1] == "") or (redis.call("get", KEYS[1]) == ARGV[1])
        then
            -- find all the keys and priorities within range
            local next_keys = redis.call("zrangebyscore", KEYS[3],
                                         ARGV[2], ARGV[3],
                                         "limit", ARGV[4], ARGV[5])
            
            if not next_keys[1] then
                return {}
            end

            local t = {}
            for i = 1, #next_keys  do
                local next_val = redis.call("hget", KEYS[2], next_keys[i])
                table.insert(t, next_keys[i])
                table.insert(t, next_val)
            end

            return t
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')
        if limit is None: limit = -1
        res = script(keys=[self._lock_name,
                           self._namespace(dict_name),
                           self._namespace(dict_name) + 'keys'],
                     args=[self._session_lock_identifier or '',
                           priority_min, priority_max, start, limit])
        if res == -1:
            raise LockError()
        split_res = {self._decode(res[i]):
                     self._decode(res[i+1])
                     for i in xrange(0, len(res)-1, 2)}
        return split_res

    def set_1to1(self, dict_name, key1, key2):
        '''Set two keys to be equal in a 1-to-1 mapping.

        Within `dict_name`, `key1` is set to `key2`, and `key2` is set
        to `key1`.

        This always requires a session lock.

        :param str dict_name: dictionary to update
        :param str key1: first key/value
        :param str key2: second key/value

        '''
        if self._session_lock_identifier is None:
            raise ProgrammerError('must acquire lock first')
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            redis.call("hset", KEYS[2], ARGV[2], ARGV[3])
            redis.call("hset", KEYS[2], ARGV[3], ARGV[2])
        else
            -- ERROR: No longer own the lock
            return -1
        end
        ''')
        res = script(keys=[self._lock_name,
                           self._namespace(dict_name)],
                     args=[self._session_lock_identifier,
                           self._encode(key1),
                           self._encode(key2)])
        if res == -1:
            raise EnvironmentError()

    def get(self, dict_name, key, default=None, include_priority=False):
        '''Get the value for a specific key in a specific dictionary.

        If `include_priority` is false (default), returns the value
        for that key, or `default` (defaults to :const:`None`) if it
        is absent.  If `include_priority` is true, returns a pair of
        the value and its priority, or of `default` and :const:`None`.

        This does not use or enforce the session lock, and is read-only,
        but inconsistent results are conceivably possible if the caller
        does not hold the lock and `include_priority` is set.

        :param str dict_name: name of dictionary to query
        :param str key: key in dictionary to query
        :param default: default value if `key` is absent
        :param bool include_priority: include score in results
        :return: value from dictionary, or pair of value and priority

        '''
        dict_name = self._namespace(dict_name)
        key = self._encode(key)
        conn = redis.Redis(connection_pool=self.pool)
        val = conn.hget(dict_name, key)
        _val = val and self._decode(val) or default
        if include_priority:
            if val:
                priority = conn.zscore(dict_name + 'keys', key)
                return _val, priority
            else:
                return _val, None
        return _val

    def set(self, dict_name, key, value, priority=None):
        '''Set a single value for a single key.

        This requires a session lock.

        :param str dict_name: name of the dictionary to update
        :param str key: key to update
        :param str value: value to assign to `key`
        :param int priority: priority score for the value (if any)

        '''
        if priority is not None:
            priorities = {key: priority}
        else:
            priorities = None
        self.update(dict_name, {key: value}, priorities=priorities)

    def delete(self, dict_name):
        '''Delete an entire dictionary.

        This operation on its own is atomic and does not require a
        session lock, but a session lock is honored.

        :param str dict_name: name of the dictionary to delete
        :raises rejester.exceptions.LockError: if called with a session
          lock, but the system does not currently have that lock; or if
          called without a session lock but something else holds it
        '''
        conn = redis.Redis(connection_pool=self.pool)
        script = conn.register_script('''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            redis.call("del", KEYS[2], KEYS[3])
            return 0
        else
            return -1
        end
        ''')
        res = script(keys=[self._lock_name,
                           self._namespace(dict_name),
                           self._namespace(dict_name) + 'keys'],
                     args=[self._session_lock_identifier])
        if res == -1:
            raise LockError()

    def direct_call(self, *args):
        '''execute a direct redis call against this Registry instances
        namespaced keys.  This is low level is should only be used for
        prototyping.

        arg[0] = redis function
        arg[1] = key --- will be namespaced before execution
        args[2:] = args to function

        :returns raw return value of function:

        Neither args nor return values are encoded/decoded
        '''
        conn = redis.Redis(connection_pool=self.pool)
        command = args[0]
        key = self._namespace(args[1])
        args = args[2:]
        func = getattr(conn, command)
        return func(key, *args)


    def increment(self, dict_name, key, value=1.0):
        '''
        increment the value stored at dict_name(key) by value
        '''
        conn = redis.Redis(connection_pool=self.pool)
        if isinstance(value, int):
            value = float(value)
            ## while redis will allow you to go from int to float, you
            ## cannot go back to using hincrby if a hash has floats in
            ## it.
            #conn.hincrby(self._namespace(dict_name), self._encode(key), value)
        if isinstance(value, float):
            conn.hincrbyfloat(self._namespace(dict_name), self._encode(key), value)
        else:
            raise TypeError('%r is not int or float' % value)




'''Redis-based distributed work manager.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Rejester is a distributed work manager, using the `Redis`_ in-memory
database as shared state.  In typical use, some number of systems run
rejester *worker* processes, which query the central database to find
new work; anything that can reach the database can submit jobs.  The
standard rejester work system can also directly deliver work units to
programs without using the dedicated worker process.  For specialized
applications there is also a globally atomic string-based priority
queue.

To use a rejester-based application:

1. Create a YAML configuration file as described below.

2. Run ``rejester mode run`` once on any system to enable rejester.

3. Run :command:`rejester_worker` on some number of systems with
   identical copies of the configuration file.  These may be anywhere,
   but must be able to reach the `Redis`_ server.  Each will start as
   many workers as the system has CPU cores.

4. Run some tool that generates rejester work units.  For instance,
   the :command:`streamcorpus_directory` tool included in
   `streamcorpus-pipeline`_ generates work units to process text
   files.

5. Run ``rejester summary`` and other sub-commands of the
   :command:`rejester` tool to get information about the job's progress.

6. Run ``rejester mode terminate`` to instruct all of the workers to
   shut down.

.. _Redis: http://redis.io/
.. _streamcorpus-pipeline: https://github.com/trec-kba/streamcorpus-pipeline

:command:`rejester` tool
========================

.. automodule:: rejester.run

:command:`rejester_worker` tool
===============================

.. automodule:: rejester.run_multi_worker

.. currentmodule:: rejester

Configuration
=============

Rejester uses :mod:`yakonfig` for its configuration.  The relevant
section of the configuration file looks like:

.. code-block:: yaml

    rejester:
      # These two options are required
      registry_addresses: [ "redis.example.com:6379" ]
      namespace: rejester

      # The following are optional, defaults shown
      app_name: rejester
      default_lifetime: 900
      enough_memory: false
      worker: fork_worker

`registry_addresses` indicates the location of the `Redis`_ server.
While this is a list, only the first value is used.  Also note that
YAML syntax requires quoting the ``host:port`` string, lest it be
interpreted as a dictionary.

`app_name` and `namespace` identify the specific application in use.
Multiple applications can share the same `Redis`_ server so long as
they have distinct namespace strings.  `app_name` is currently fixed
at ``rejester`` and setting a different value has no effect.

`default_lifetime` indicates how long a job is allowed to run before
it must call :meth:`rejester._task_master.WorkUnit.update`.  If a
job runs beyond this timeout it will return from the "pending" list to
the "available" list, and another worker may start working on it.

When a worker requests a job, if the system does not have enough memory
to satisfy a given work spec's ``min_gb`` request, that work spec is
skipped.  Setting `enough_memory` to true overrides this check.

`worker` specifies the worker implementation to use in
:command:`rejester_worker`.  Valid options are ``fork_worker`` or
``multi_worker``.  ``fork_worker`` has additional configuration
options; see :class:`~rejester.workers.ForkWorker` for details.
``multi_worker`` is less stable and never allows jobs to time out, but
will start a set of jobs more quickly.

Task system
===========

Core API
--------

.. autoclass:: TaskMaster
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: WorkUnit
   :members:
   :undoc-members:
   :show-inheritance:

Workers
-------

.. autoclass:: Worker
   :members:
   :undoc-members:
   :show-inheritance:

.. automodule:: rejester.workers
.. currentmodule:: rejester

Implementation Details
----------------------

.. autoclass:: Registry
   :members:
   :undoc-members:
   :show-inheritance:

Priority queue system
=====================

.. autoclass:: RejesterQueue
   :members:
   :undoc-members:
   :show-inheritance:

Exceptions
==========

.. automodule:: rejester.exceptions
   :members:
   :undoc-members:
   :show-inheritance:



Example: StreamCorpus Simple Filter Stage
=========================================

Rejester makes it easy to run large batches of small jobs, such as a
couple million runs of tagging batches of 500 documents with an NER
tagger or filtering to smaller sets of documents.

As a simple example, we illustrate how to write a filter function as
an external stage and run it in AWS EC2.

:mod:`streamcorpus_pipeline` has several built-in filters `source in github
<https://github.com/trec-kba/streamcorpus-pipeline/blob/master/streamcorpus_pipeline/_filters.py>`_.
You can create your own as external stages.  For example, see this `exact name match filter
<https://github.com/trec-kba/streamcorpus-pipeline/blob/master/examples/filters_exact_match.py>`_.

.. code-block:: py

    ## use the newer regex engine
    import regex as re
    import string

    ## make a unicode translation table to converts all punctuation to white space
    strip_punctuation = dict((ord(char), u" ") for char in string.punctuation)

    white_space_re = re.compile("\s+")

    def strip_string(s):
        'strips punctuation and repeated whitespace from unicode strings'
        return white_space_re.sub(" ", s.translate(strip_punctuation).lower())


    class filter_exact_match(object):
        'trivial string matcher using simple normalization and regex'

        config_name = 'filter_exact_match'

        def __init__(self, config):
            path = config.get('match_strings_path')
            match_strings = open(path).read().splitlines()
            match_strings = map(strip_string, map(unicode, match_strings))
            self.matcher = re.compile("(.|\\n)*?(%s)" % '|'.join(match_strings), re.I)

        def __call__(self, si, context):
            'only pass StreamItems that match'
            if self.matcher.match(strip_string(si.body.clean_visible.decode('utf-8'))):
                return si


    ## this is how streamcorpus_pipeline finds the stage
    Stages = {'filter_exact_match': filter_exact_match}


To run this in rejester, you need to setup a `redis server
<http://redis.io/>`_ (use version 2.8 or newer), and put the hostname
in your yaml configuration file:

.. code-block:: yaml

    logging:
      root:
        level: INFO

    rejester:
      namespace: my_kba_filtering
      app_name: rejester
      registry_addresses: ["redis.example.com:6379"]

    streamcorpus_pipeline:
      ## "." means current working directory
      root_path: .

      tmp_dir_path: tmp
      cleanup_tmp_files: true

      external_stages_path: examples/filter_exact_match.py

      reader: from_s3_chunks

      incremental_transforms:
        ## remove all StreamItems that do not exactly match
        - filter_exact_match

      batch_transforms: []

      filter_exact_match:
        ## names ending in "_path" will be made absolute relative to the root_path
        match_strings_path: my_match_strings.txt

      writers: [to_s3_chunks]

      from_s3_chunks:
        ## put paths to your own keys here; these files must be just the
        ## access_key_id and secret_access_key strings without newlines
        aws_access_key_id_path:     /data/trec-kba/keys/trec-aws-s3.aws_access_key_id
        aws_secret_access_key_path: /data/trec-kba/keys/trec-aws-s3.aws_secret_access_key

        ## this is the location of the NIST's StreamCorpus
        bucket: aws-publicdatasets
        s3_path_prefix: trec/kba/kba-streamcorpus-2014-v0_3_0

        tries: 10
        input_format: streamitem
        streamcorpus_version: v0_3_0

        ## you need this key if you are processing NIST's StreamCorpus,
        ## which is encrypted
        gpg_decryption_key_path: /data/trec-kba/keys/trec-kba-rsa.gpg-key.private

      to_s3_chunks:

        ## put your own bucket and paths to your own keys here; these
        ## files must be just the access_key_id and secret_access_key
        ## strings without newlines
        aws_access_key_id_path:     /data/trec-kba/keys/trec-aws-s3.aws_access_key_id
        aws_secret_access_key_path: /data/trec-kba/keys/trec-aws-s3.aws_secret_access_key

        bucket: aws-publicdatasets
        s3_path_prefix: trec/kba/kba-streamcorpus-2014-v0_3_0-to-delete

        output_name: "%(date_hour)s/%(source)s-%(num)d-%(input_md5)s-%(md5)s"
        tries: 10
        cleanup_tmp_files: true

        ## you only need these keys if you are encrypting the data that
        ## you are putting into your bucket; you need both if you require
        ## verify_via_http, which fetches and decrypts to validate what
        ## was saved.
        gpg_decryption_key_path: /data/trec-kba/keys/trec-kba-rsa.gpg-key.private
        gpg_encryption_key_path: /data/trec-kba/keys/trec-kba-rsa.gpg-key.pub
        gpg_recipient: trec-kba

        verify_via_http: true


Then, you can run the following commands to put tasks into the rejester queue:

.. code-block:: bash

    ## populate the task queue with jobs
    streamcorpus_directory -c examples/streamcorpus-2014-v0_3_0-exact-match-example.yaml --file-lists list_of_s3_paths.txt

    ## try running one, to make sure it works locally
    rejester -c examples/streamcorpus-2014-v0_3_0-exact-match-example.yaml run_one

    ## launch a MultiWorker to use all the CPUs on this machine:
    rejester -c examples/streamcorpus-2014-v0_3_0-exact-match-example.yaml run_worker

    ## check on the status of your jobs
    rejester -c examples/streamcorpus-2014-v0_3_0-exact-match-example.yaml summary


If you are interested in SaltStack states for spinning up machines to
do this, reach out on streamcorpus@googlegroups.com

'''
from __future__ import absolute_import
from rejester.config import config_name, default_config, add_arguments, \
    runtime_keys, discover_config, check_config
from rejester._task_master import TaskMaster, WorkUnit, Worker, build_task_master
from rejester._queue import RejesterQueue
from rejester._registry import Registry
